import typing
import weakref


class WeakFunctionReferences:
    def __init__(self):
        self.references = {}

    def add(self, function: typing.Callable):
        callback_id = hash(function) & 0xffffffffffffff
        if callback_id in self.references:
            return callback_id
        self.references[callback_id] = weakref.WeakMethod(function, self._create_delete_callback(callback_id))
        return callback_id

    def remove(self, function: typing.Callable):
        del self.references[hash(function) & 0xffffffffffffff]

    def get(self, callback_id: int) -> typing.Callable:
        return self.references[callback_id]()

    def get_all(self):
        return [hr for hr in [wr() for wr in self.references.values()] if hr is not None]

    def _create_delete_callback(self, callback_id: int) -> typing.Callable:
        # noinspection PyUnusedLocal
        def f(wr):
            try:
                del self.references[callback_id]
            except:
                pass
        return f



import json, sys
import traceback
import logging
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWebChannel import QWebChannel
from PyQt5.QtWidgets import *
from PyQt5.QtWebEngineWidgets import *
from pyhtmlgui import Observable

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)  # enable highdpi scaling
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)  # use highdpi icons

AppKit = None
if sys.platform == "darwin":
    try:
        import AppKit
    except:
        logging.warning("AppKit not available, hiding dock icon will not be available. Run 'pip install pyobjc' to install ")
    NSApplicationActivationPolicyRegular = 0
    NSApplicationActivationPolicyAccessory = 1
    NSApplicationActivationPolicyProhibited = 2


class PyHtmlQtApp(QApplication):
    def __init__(self, icon_path= None ):
        super(PyHtmlQtApp, self).__init__([])
        self.setQuitOnLastWindowClosed(False)
        self._icon_cache = {}
        self._current_icon = None
        self.set_icon(icon_path)

        self.on_about_to_quit_event = Observable()
        self.on_activated_event = Observable()
        self.aboutToQuit.connect(self.on_about_to_quit_event.notify_observers)

    def run(self):
        self.exec_()

    def stop(self):
        self.quit()

    def set_icon(self, path):
        if path not in self._icon_cache:
            self._icon_cache[path] = QIcon(path)
        self._current_icon = self._icon_cache[path]
        self.setWindowIcon(self._icon_cache[path])

    def event(self, e):
        if e.type() == QEvent.ApplicationActivate and e.spontaneous() is True:
            self.on_activated_event.notify_observers()
        return QApplication.event(self, e)

    def hide_osx_dock(self):
        if AppKit is not None:
            AppKit.NSApp.setActivationPolicy_(NSApplicationActivationPolicyProhibited)
        else:
            logging.warning("AppKit not available, hide_osx_dock() is not available")

    def show_osx_dock(self):
        if AppKit is not None:
            AppKit.NSApp.setActivationPolicy_(NSApplicationActivationPolicyRegular)
            self.setWindowIcon(self._current_icon)
        else:
            logging.warning("AppKit not available, show_osx_dock() is not available")


class GenericTray():
    def __init__(self, pyHtmlQtApp, icon_path):
        self._pyHtmlQtApp = pyHtmlQtApp
        self._tray = QSystemTrayIcon(self._pyHtmlQtApp)
        self._icon_cache = {}
        self.set_icon(icon_path)
        self._tray.setVisible(True)

        self._menu = QMenu()
        self._menu._subitems = {}

        self._tray.setContextMenu(self._menu)

        self.on_left_clicked = Observable()
        self.on_right_clicked = Observable()
        self.on_closed_event = Observable()
        self.on_show_event = Observable()

        self._tray.activated.connect(self._right_or_left_click)
        self._menu.aboutToHide.connect(self.on_closed_event.notify_observers) # https://doc.qt.io/qt-5/macos-issues.html#menu-actions
        self._menu.aboutToShow.connect(self.on_show_event.notify_observers)

    def show(self):
        self._tray.show()

    def hide(self):
        self._tray.hide()

    def close(self):
        self._tray.hide()

    def set_icon(self, path):
        if path not in self._icon_cache:
            self._icon_cache[path] = QIcon(path)
        self._tray.setIcon(self._icon_cache[path])

    def _right_or_left_click(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.Trigger:
            self.on_left_clicked.notify_observers()
        elif reason == QSystemTrayIcon.ActivationReason.Context:
            self.on_right_clicked.notify_observers()


class PyHtmlQtSimpleTray(GenericTray):
    def addAction(self, name_or_names, target):
        if type(name_or_names) == str:
            name_or_names = [ name_or_names]
        submenu = self._get_submenu(name_or_names[:-1])
        action = QAction(parent=submenu, text=name_or_names[-1])
        action.triggered.connect(target)
        submenu.addAction(action)

    def addSeparator(self, names = []):
        submenu = self._get_submenu(names)
        submenu.addSeparator()

    def _get_submenu(self, names):
        parent = self._menu
        for name in names:
            if name in parent._subitems:
                parent = parent._subitems[name]
            else:
                subitem = parent.addMenu(name)
                subitem._subitems = {}
                parent._subitems[name] = subitem
                parent = subitem
        return parent


class PyHtmlQtTray(GenericTray):
    def __init__(self, pyHtmlQtApp, url, size, icon_path, keep_connected_on_close = True):
        super().__init__(pyHtmlQtApp, icon_path)
        self._webWidget = PyHtmlWebWidget(url, size=size)
        self._trayAction = QWidgetAction(self._tray)
        self._trayAction.setDefaultWidget(self._webWidget)
        self._menu.addAction(self._trayAction)
        if keep_connected_on_close is False:
            self.on_show_event.attach_observer(self._webWidget.load_page)
            self.on_closed_event.attach_observer(self._webWidget.unload_page)
        else:
            self._webWidget.load_page() # load page now if it should stay active in background

    def addJavascriptFunction(self, name, target):
        self._webWidget.addJavascriptFunction(name, target)

    def removeJavascriptFunction(self, name):
        self._webWidget.removeJavascriptFunction(name)

    def runJavascript(self, javascript, callback=None):
        self._webWidget.runJavascript(javascript, callback)


class PyHtmlQtWindow():
    def __init__(self, pyHtmlQtApp, url, size, title, icon_path=None, keep_connected_on_close = False, keep_connected_on_minimize = True):
        self._pyHtmlQtApp = pyHtmlQtApp

        self.on_closed_event = Observable()
        self.on_show_event = Observable()
        self.on_minimized_event = Observable()

        self._webWidget = PyHtmlWebWidget(url)
        self._qMainWindow = ExtendedQMainWindow(self) # you could also let this class directly subclass QMainWindow, but then PyHtmlQtWindow will expose a s*load of confusing qt functions from QMainWindow
        self._qMainWindow.resize(size[0], size[1])
        self._qMainWindow.setFocusPolicy(Qt.StrongFocus)
        self._qMainWindow.setWindowTitle(title)
        self._qMainWindow.setCentralWidget(self._webWidget)
        self._webWidget.setParent(self._qMainWindow)
        self._menuBar = QMenuBar()
        self._menuBar._subitems = {} # qt also keep track of this, but this is easier for the addMenuButton and addMenuSeparator functions
        self._qMainWindow.setMenuBar(self._menuBar)

        self._icon_cache = {}
        self.set_icon(icon_path)

        if keep_connected_on_close is False:
            self.on_closed_event.attach_observer(self._webWidget.unload_page)
        if keep_connected_on_minimize is False:
            self.on_minimized_event.attach_observer(self._webWidget.unload_page)
        self.on_show_event.attach_observer(self._webWidget.load_page)

    def show(self):
        self._qMainWindow.showNormal()
        self._qMainWindow.activateWindow()
        self._qMainWindow.raise_()

    def hide(self):
        self._qMainWindow.hide()

    def minimize(self):
        self._qMainWindow.showMinimized()

    def close(self):
        self._qMainWindow.close()

    def set_title(self, title):
        self._qMainWindow.setWindowTitle(title)

    def set_icon(self, path):
        if path not in self._icon_cache:
            self._icon_cache[path] = QIcon(path)
        self._qMainWindow.setWindowIcon(self._icon_cache[path])

    def addMenuButton(self, name_or_names, target):
        if type(name_or_names) == str:
            name_or_names = [ name_or_names]
        submenu = self._get_submenu(name_or_names[:-1])
        action = QAction(text=name_or_names[-1], parent=submenu)
        action.triggered.connect(target)
        submenu.addAction(action)

    def addMenuSeparator(self, names = []):
        submenu = self._get_submenu(names)
        submenu.addSeparator()

    def addJavascriptFunction(self, name, target):
        self._webWidget.addJavascriptFunction(name, target)

    def removeJavascriptFunction(self, name):
        self._webWidget.removeJavascriptFunction(name)

    def runJavascript(self, javascript, callback=None):
        self._webWidget.runJavascript(javascript, callback)

    def _get_submenu(self, names):
        parent = self._menuBar
        for name in names:
            if name in parent._subitems:
                parent = parent._subitems[name]
            else:
                subitem = parent.addMenu(name)
                subitem._subitems = {}
                parent._subitems[name] = subitem
                parent = subitem
        return parent


class PyHtmlWebWidget(QWidget):
    def __init__(self, url, size=None):
        super().__init__()
        self.url  = url
        self._current_url = None

        self.web = QWebEngineView(parent=self)
        if size is not None:
            self.web.setFixedSize(size[0], size[1])

        self.profile = QWebEngineProfile(parent= self) # create a seperate profile for earch webengine, otherwise QWebEngineViews will load only the first loaded url in all view.
        self.web.setPage(QWebEnginePage(self.profile, self.web))
        self.web.load(QUrl("about:blank"))

        self.channel = QWebChannel(parent=self.web)
        self._functionHandler = JavascriptFunctionHandler()
        self.channel.registerObject('pyhtmlapp', self._functionHandler)
        self.web.page().setWebChannel(self.channel)

        self.layout = QGridLayout()
        self.layout.addWidget(self.web, 0, 0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.setRowStretch(0, 0)
        self.setLayout(self.layout)

    def load_page(self):
        if self._current_url != self.url:
            self._current_url = self.url
            self.web.load(QUrl(self.url))

    def unload_page(self):
        if self._current_url != None:
            self._current_url = None
            self.web.load(QUrl("about:blank"))

    def addJavascriptFunction(self, name, target):
        self._functionHandler.add(name, target)

    def removeJavascriptFunction(self, name):
        self._functionHandler.remove(name)

    def runJavascript(self, javascript, callback=None):
        if callback is None:
            self.web.page().runJavaScript(javascript)
        else:
            self.web.page().runJavaScript(javascript, callback)


class ExtendedQMainWindow(QMainWindow):
    def __init__(self, pyHtmlQtWindow):
        self.pyHtmlQtWindow = pyHtmlQtWindow
        super().__init__()

    def closeEvent(self, event) -> None:
        self.pyHtmlQtWindow.on_closed_event.notify_observers()

    def showEvent(self, event) -> None:
        self.pyHtmlQtWindow.on_show_event.notify_observers()

    def hideEvent(self, event) -> None:
        self.pyHtmlQtWindow.on_minimized_event.notify_observers()


class JavascriptFunctionHandler(QObject):
    def __init__(self):
        super().__init__()
        self._actions = {}

    @pyqtSlot(str, str, result=str)
    def call(self, action_name, arguments):
        if action_name not in self._actions:
            return json.dumps({"exception": "Unknown function pyhtmlapp.%s, did you forget to call addJavascriptFunction ?" % action_name})
        try:
            arguments = json.loads(arguments) # do in multiple lines, so if something failes, we see where.
            action = self._actions[action_name]
            result = action(*arguments)
            return json.dumps({"result" : result})
        except Exception as e:
            msg = "Exception while calling python from javascript function 'pyhtmlapp.%s'\n%s" %  (action_name, traceback.format_exc())
            return json.dumps({"exception": msg })

    def add(self, name, target):
        self._actions[name] = target

    def remove(self, name):
        self._actions.pop(name)

    def get(self, name):
        return  self._actions[name]


from __future__ import annotations

import threading
import time
import typing
import jinja2
import weakref
import traceback
import re
import os
import sys
import inspect
import json
import logging
import queue
import importlib
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyhtmlgui.pyhtmlgui import PyHtmlGui
    from pyhtmlgui.view.pyhtmlview import PyHtmlView
from .lib import WeakFunctionReferences

class PyHtmlGuiInstance:
    def __init__(self, parent: PyHtmlGui, app_instance: object, view_class: typing.Type[PyHtmlView], _on_dom_ready_callback):
        self._parent = parent
        self._on_dom_ready_callback = _on_dom_ready_callback
        self._websocket_connections = []
        self._children = weakref.WeakSet()
        self._template_env = jinja2.Environment(loader=parent.template_loader, autoescape=jinja2.select_autoescape())
        self._template_cache = {}
        self._call_number = 0
        self._function_references = WeakFunctionReferences()
        self._template_env.globals['_create_py_function_reference'] = self._create_function_reference
        self.pending_js_results = {}
        self._polling_children = {}
        self._polling_thread = None
        self._view = view_class(app_instance, self)

    @property
    def connections_count(self) -> int:
        return len(self._websocket_connections)

    def update(self) -> None:
        self.call_javascript("pyhtmlgui.update_element", ["pyHtmlGuiBody", self._view.render()], skip_results=True)

    def set_visible(self, visible: bool) -> None:
        """
        Set component and childens visibility, components that are not visible get their events detached
        """
        if visible is False:
            for child in self._children:
                try:
                    if child.is_visible is True:
                        child.set_visible(False)
                except Exception as e:
                    pass
        else:
            self.update()

    def call_javascript(self, js_function_name: str, args: list = None, skip_results: bool = False):
        """
        Call javascript function in frontend.
        :param js_function_name: Name of javascript function
        :param args: Arguments for js function
        :param skip_results: Don't receive results, give some slight performance improvement if we don't wait for result
        """

        self._call_number  += 1
        call_id = self._call_number

        try:  #remove old function references, this may not be needed, because it should be removed on return from python, but in case of error it may stay forever otherwise
            self.pending_js_results[call_id - 1000].removed_by_parent() # will call outstanding callbacks, collecting as many results as have been received at this point
            del self.pending_js_results[call_id - 1000] # no more that 1000 calls should stay around
        except:
            pass

        javascript_call_object = {'call': call_id, 'name': js_function_name, 'args': args if args is not None else []}
        javascript_call_result = None
        if skip_results is False:
            javascript_call_result = JavascriptCallResult(self, call_id, len(self._websocket_connections) )
            self.pending_js_results[call_id] = javascript_call_result
        else:
            javascript_call_object["skip_results"] = True

        data = json.dumps(javascript_call_object, default=lambda o: None)
        for websocket_connection in [w for w in self._websocket_connections]:
            websocket_connection.send(data)
        return javascript_call_result

    def process(self, ws) -> None:
        websocket_connection = WebsocketConnection(ws, self)
        self._websocket_connections.append(websocket_connection)
        websocket_connection.receive_loop()
        self._websocket_connections.remove(websocket_connection)
        if len(self._websocket_connections) == 0:
            self.set_visible(False)

    def _create_function_reference(self, function: typing.Union[typing.Callable, jinja2.runtime.Undefined]) -> str:
        if type(function) == jinja2.runtime.Undefined:
            # noinspection PyProtectedMember
            raise Exception("Undefined python method in script: '%s'" % function._undefined_name)
        return self._function_references.add(function)

    def get_template(self, item: PyHtmlView, force_reload: bool = False):
        """
        Receive template associated to item
        :param item: The view Object
        :param force_reload: Force reloading of template
        :return:
        """
        if force_reload is False:
            try:
                return self._template_cache[item.__class__.__name__]
            except KeyError:
                pass

        if item.__class__.TEMPLATE_FILE is not None:  # load from file
            file_to_monitor = self._template_env.get_template(item.TEMPLATE_FILE).filename
            string_to_render = open(file_to_monitor, "r").read()
        else:  # load from class
            if self._parent.auto_reload is False:
                string_to_render = item.TEMPLATE_STR
                file_to_monitor = None
            else:
                module_name = item.__module__
                if module_name is None or module_name == str.__class__.__module__:
                    module_fullname = item.__class__.__name__  # Avoid reporting __builtin__
                else:
                    module_fullname = module_name + '.' + item.__class__.__name__

                try:
                    file_to_monitor = os.path.abspath(inspect.getfile(item.__class__))
                except Exception:  # in case its in main. this may be a bug in inspect
                    file_to_monitor = os.path.abspath(sys.argv[0])

                if module_name == "__main__":
                    name = os.path.splitext(os.path.basename(file_to_monitor))[0]
                    module = __import__(name)
                    importlib.reload(module)  # reload should work on non complex objects in __main__, but not for more
                    for comp in module_fullname.split(".")[1:]:
                        module = getattr(module, comp)
                else:
                    loader = importlib.machinery.SourceFileLoader(module_name, file_to_monitor)
                    # noinspection PyUnresolvedReferences
                    spec = importlib.util.spec_from_loader(loader.name, loader)
                    # noinspection PyUnresolvedReferences
                    module = importlib.util.module_from_spec(spec)
                    loader.exec_module(module)
                    module = getattr(module, module_fullname.split(".")[-1])
                string_to_render = module.TEMPLATE_STR

        if self._parent.auto_reload is True:
            self._parent._add_file_to_monitor(file_to_monitor, item.__class__.__name__)

        string_to_render = self._prepare_template(string_to_render)

        try:
            self._template_cache[item.__class__.__name__] = self._template_env.from_string(string_to_render)
        except Exception as e:
            msg = "Failed to load Template "
            if item.TEMPLATE_FILE is not None:
                msg += "from File '%s': " % item.TEMPLATE_FILE
            else:
                msg += "from Class '%s': " % item.__class__.__name__
            msg += " %s" % e
            raise Exception(msg)

        return self._template_cache[item.__class__.__name__]

    @staticmethod
    def _prepare_template(template: str) -> str:
        """
            Replace onclick="pyview.my_function(arg1,arg2)"
            with    onclick="pyhtmlgui.call({{_create_py_function_reference(pyview.my_function)}}, arg1, arg2)
        """
        parts = re.split('({{|}}|{%|%})', template)
        index = 0
        while index < len(parts):
            if parts[index] == "{{" or parts[index] == "{%" :
                parts[index] = "%s%s%s" % (parts[index] , parts[index+1], parts[index+2])
                del parts[index+ 1]
                del parts[index+ 1]
            index += 1

        new_parts = []
        for i, part in enumerate(parts):
            if part.startswith("{{") or part.startswith("{%") or part.find("pyview.") == -1:
                new_parts.append(part)
            else:
                # noinspection RegExpSingleCharAlternation
                subparts = re.split(r'(>| |\(|=|\"|\'|\n|\r|\t|;)(pyview.[a-zA-Z0-9_.]+\()', part)
                for x, subpart in enumerate(subparts):
                    if subpart.startswith("pyview."):
                        subpart = subpart.replace("(", ")}}, ", 1)
                        subparts[x] = "pyhtmlgui.call({{_create_py_function_reference(%s" % subpart
                for sp in subparts:
                    new_parts.append(sp)
        return "".join(new_parts).replace(r'\pyview.', 'pyview.')

    def clear_template_cache(self, classnames: str = None) -> None:
        if classnames is None:
            self._template_cache = {}
        else:
            for classname in classnames:
                try:
                    del self._template_cache[classname]
                except Exception:
                    pass

    def _add_child(self, child: PyHtmlView) -> None:
        self._children.add(child)

    def _remove_child(self, child: PyHtmlView) -> None:
        self._children.remove(child)

    def _add_polling_child(self, child: PyHtmlView) -> None:
        if self._polling_thread == None:
            self._polling_thread = threading.Thread(target=self._poll_loop, daemon=True)
            self._polling_thread.start()
        try:
            if child in self._polling_children[child._autoupdate_interval]:
                return
        except:
            pass
        self._remove_polling_child(child) # in case interval has changed and child already exists in some queue
        if child._autoupdate_interval not in self._polling_children:
            self._polling_children[child._autoupdate_interval] = weakref.WeakSet()
        self._polling_children[child._autoupdate_interval].add(child)

    def _remove_polling_child(self, child):
        for key in self._polling_children.keys():
            try:
                self._polling_children[key].remove(child)
            except:
                pass
            if len(self._polling_children[key]) == 0:
                del self._polling_children[key]

    def _poll_loop(self):
        i = 0
        while True:
            now = time.time()
            for key in self._polling_children.keys():
                interval = int(key)
                if i % interval == 0:
                    for child in self._polling_children[key]:
                        if child.is_visible and child._last_rendered + interval/2.0 < now:
                            child.update()
            i += 1
            time.sleep(1)


class WebsocketConnection:
    def __init__(self, ws, pyHtmlGuiInstance):
        self.ws = ws
        self.parent_instance = pyHtmlGuiInstance
        self.active = True
        self.send_queue = queue.Queue(maxsize=1000)
        self._send_t = threading.Thread(target=self._send_loop, daemon=True)
        self._send_t.start()

    def receive_loop(self):
        while self.active is True:
            try:
                msg = self.ws.receive()
            except:
                msg = None
            if msg is None:
                break
            try:
                self._process_received_message(json.loads(msg))
            except:
                continue
        try:
            self.ws.close()
        except:
            pass
        self.active = False
        if self.send_queue is not None:
            self.send_queue.put(None)

    def send(self, message):
        if self.send_queue is not None:
            self.send_queue.put(message)

    def _process_received_message(self, message):
        if 'call' in message:
            function_name = "Function not found"
            args = "  "
            return_val = None
            try:
                if message['name'] == "call_python_function_with_args":
                    functioncall_id = message['args'][0]
                    args = message['args'][1]
                    function = self.parent_instance._function_references.get(functioncall_id)
                    # noinspection PyUnresolvedReferences
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function(*args)

                elif message['name'] == "call_python_function":
                    functioncall_id = message['args'][0]
                    function = self.parent_instance._function_references.get(functioncall_id)
                    # noinspection PyUnresolvedReferences
                    function_name = "%s.%s" % (function.__self__.__class__.__name__, function.__name__)
                    return_val = function()

                elif message['name'] == "frontend_ready":
                    function_name = "frontend_ready"
                    self.parent_instance.update()
                    self.parent_instance._on_dom_ready_callback()

                elif message['name'] == "ping":
                    pass

                else:
                    logging.error("unknown python function '%s'" % message['name'])

            except Exception:
                tb = traceback.format_exc()
                msg = " Exception in: %s(%s)\n" % (function_name, ("%s" % args)[1:-1])
                msg += " %s" % tb.replace("\n", "\n  ").strip()
                self.parent_instance.call_javascript("pyhtmlgui.debug_msg", [msg], skip_results=True)
                logging.error(msg)
                return_val = None

            if not ("skip_results" in message and message["skip_results"] is True):
                data = json.dumps({'return': message['call'], 'value': return_val}, default=lambda o: None)
                self.send(data)

        elif 'return' in message:
            call_id = message['return']
            del message['return']  # remove internal id from result before passing to next level
            try:
                self.parent_instance.pending_js_results[call_id].result_received(message)
            except:
                pass
        else:
            logging.error('Invalid message received: %s' % message)

    def _send_loop(self):
        while self.active is True:
            message = self.send_queue.get()
            if message is None:
                break
            try:
                self.ws.send(message)
            except:
                pass
        try:
            self.ws.close()
        except:
            pass
        self.active = False
        self.send_queue = None


class JavascriptCallResult:
    def __init__(self, pyHtmlGuiInstance, call_id, nr_of_expected_results):
        self.instance = pyHtmlGuiInstance
        self.call_id = call_id
        self._results_expected = nr_of_expected_results
        self.results = []
        self._callback = None
        self._all_results_received_event = threading.Event()

    def result_received(self, result):
        self.results.append(result)
        if len(self.results) == self._results_expected:
            if self._callback is not None:
                self._callback(self.results)
            self._all_results_received_event.set()
            self._clear_parent_reference()

    def _get_results_blocking(self):
        self._all_results_received_event.wait(60)
        errors = [result["error"] for result in self.results if "error" in result]
        if len(errors) > 0:
            msg = "%s of %s connected frontends returned an error\n" % (len(errors), len(self.results))
            msg += "\n".join(errors)
            raise Exception(msg)
        return [result["value"] for result in self.results]

    def removed_by_parent(self):
        self._all_results_received_event.set()

    def _clear_parent_reference(self):
        try:
            del self.instance.pending_js_results[self.call_id]  # remove pending reference
        except:
            pass

    def __call__(self, callback: typing.Union[typing.Callable, None]  = None):
        if callback is None:
            r = self._get_results_blocking()
            self._clear_parent_reference()
            return r
        else:
            if self._all_results_received_event.is_set() is True:
                callback(self._get_results_blocking())
                self._clear_parent_reference()
            else:
                self._callback = callback



from __future__ import annotations

import json
import os
import logging
import sys
import threading
import time
import traceback
import uuid
import webbrowser

import jinja2
import typing
from werkzeug.serving import make_server

from .pyhtmlguiInstance import PyHtmlGuiInstance
from .view import PyHtmlView
import flask, flask_sock

class PyHtmlGui:
    def __init__(self,
                 app_instance     : object,
                 view_class       : typing.Type[PyHtmlView],
                 static_dir       : str             = None,
                 template_dir     : str             = None,
                 base_template    : str             = "pyHtmlGuiBase.html",
                 on_view_connected: typing.Callable = None,
                 on_view_disconnected: typing.Callable = None,
                 size             : tuple[int, int] = None,
                 position         : tuple[int, int] = None,
                 listen_host      : str             = "127.0.0.1",
                 listen_port      : int             = 0,
                 shared_secret    : str             = None,
                 auto_reload      : bool            = False,
                 single_instance  : bool            = True
                 ) -> None:
        """
        :param app_instance: Some object (eg. main program class instance), passed to view_class as obj on launch
        :type  app_instance: object
        :param view_class: A class that Inherits from PyHtmlView
        :param static_dir: Static files, css, img go here
        :param template_dir: main.html and other html goes here
        :param base_template: pyHtmlGuiBase in pyHtmlGui/assets/templates, or custom file in app templates dir
        :param size: window size
        :param on_view_connected: callback is called when a frontend connects via websocket,
                                        arguments passed: "nr of view instances", "nr of websocket connections"
        :param on_view_disconnected: callback is called when a frontend websocket is disconnected
                                        arguments passed: "nr of view instances", "nr of websocket connections"
        :param position: window position
        :param listen_host:
        :param listen_port: Default listen port, of 0 for automatic
        :param shared_secret: use "" to automatically generate a uid internally, None to disable token
        :param auto_reload: for development, monitor files and reload while app is running
        :param single_instance: create only one instance and share it between all connected websockets.
                                This is the default, so one instance of view_class is shared by all connected frontends
        """

        self._endpoints = {}

        self.add_endpoint(
            name          = "",
            app_instance  = app_instance,
            view_class    = view_class,
            base_template = base_template,
            on_view_connected   = on_view_connected,
            on_view_disconnected= on_view_disconnected,
            single_instance     = single_instance,
            size     = size,
            position = position,
        )

        if shared_secret == "":
            self.shared_secret = "%s" % uuid.uuid4()
        elif shared_secret is None:
            self.shared_secret = None
        else:
            self.shared_secret = shared_secret

        self.listen_host = listen_host
        self.listen_port = listen_port
        self.auto_reload = auto_reload

        self.static_dir = None if static_dir is None else os.path.abspath(static_dir)
        self.template_dir = None if template_dir is None else os.path.abspath(template_dir)

        if getattr(sys, 'frozen', False) is True:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            self._internal_template_dir = os.path.join(sys._MEIPASS, "pyhtmlgui", "templates")
            self.auto_reload = False
        else:
            self._internal_template_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")
        if self.template_dir is not None and not os.path.exists(self.template_dir):
            raise Exception("Template dir '%s' not found" % self.template_dir)
        if self.static_dir is not None and not os.path.exists(self.static_dir):
            raise Exception("Static dir '%s' not found" % self.static_dir)

        self.template_loader = jinja2.FileSystemLoader(searchpath=[self._internal_template_dir] if self.template_dir is None else [self._internal_template_dir, self.template_dir])
        self._template_env = jinja2.Environment(loader=self.template_loader, autoescape=jinja2.select_autoescape())

        self._token_cookie = "%s" % uuid.uuid4()
        self._token_csrf = "%s" % uuid.uuid4()

        self.flaskApp = flask.Flask("pyhtmlgui", static_folder=None, template_folder=None)

        self.flaskWebsocket =  flask_sock.Sock()
        self.flaskApp.route("/")(self._main_html)
        self.flaskApp.route("/<endpoint>")(self._main_html)
        if self.static_dir is not None:
            self.flaskApp.route("/static/<path:path>")(self._static)
        self.flaskWebsocket.route("/ws")(self._websocket)
        self.flaskWebsocket.init_app(self.flaskApp)

        self._server = WebsocketServerThread(self.flaskApp, self.listen_host, self.listen_port)

        self._file_monitoring = {}
        if self.auto_reload is True:
            t = threading.Thread(target=self._monitoring_thread, daemon=True)
            t.start()

    def start(self, show_frontend: bool = False, block: bool = True) -> None:
        self._server.start()
        print("PyHtmlGui server started on http://%s:%s" % (self.listen_host, self._server.get_port()))
        if show_frontend is True:
            self.show()
        if block is True:
            self.join()

    def stop(self) -> None:
        try:
            self._server.shutdown()
        except Exception:
            pass

    def join(self):
        self._server.join()

    def show(self) -> None:
        for endpoint in self._endpoints.values():
            endpoint.show()

    def add_endpoint(self, app_instance: object, view_class: typing.Type[PyHtmlView], name: str = "", base_template: str = "pyHtmlGuiBase.html",
         on_view_connected: typing.Callable = None, on_view_disconnected: typing.Callable = None,
         single_instance: bool = True, size: tuple[int, int] = None, position: tuple[int, int] = None ):
        if name in self._endpoints:
            raise Exception("Endpoint named '%s' already exists" % name)
        self._endpoints[name] = PyHtmlGuiEndpoint(
            parent        = self,
            app_instance  = app_instance,
            name          = name,
            view_class    = view_class,
            base_template = base_template,
            on_view_connected    = on_view_connected,
            on_view_disconnected = on_view_disconnected,
            single_instance      = single_instance,
            size     = size,
            position = position
        )

    def get_url(self, endpoint = ""):
        if endpoint in self._endpoints:
            return self._endpoints[endpoint].get_url()
        return None

    # /
    def _main_html(self, endpoint = ""):
        if self.shared_secret is not None and flask.request.args.get('token') != self.shared_secret:
            return flask.abort(403)
        if endpoint not in self._endpoints:
            return flask.abort(404)
        template = self._template_env.get_template(self._endpoints[endpoint].base_template)
        response = flask.Response(template.render({
            "csrf_token"    : self._token_csrf,
            "start_size"    : json.dumps(self._endpoints[endpoint].size),
            "start_position": json.dumps(self._endpoints[endpoint].position),
        }))
        response.headers['Cache-Control'] = 'no-store'
        response.set_cookie("token", self._token_cookie)
        response.set_cookie("endpoint", endpoint)
        return response

    # /static/<path>
    def _static(self, path: str):
        if flask.request.cookies.get('token') != self._token_cookie:
            return flask.abort(403)
        if self.static_dir is None:
            return flask.abort(404)
        response = flask.helpers.send_from_directory( self.static_dir, path, max_age=36000)
        filetype = path.split(".")[-1].lower()
        if filetype == "js":
            response.headers["Content-Type"] =  "application/javascript"
        elif filetype in ["html", "css"]:
            response.headers["Content-Type"] =  "text/%s" % filetype
        elif filetype in ["jpeg", "jpg"]:
            response.headers["Content-Type"] =  "image/jpeg"
        elif filetype in ["png", "gif", "webp", "svg"]:
            response.headers["Content-Type"] =  "image/%s" % filetype

        response.headers["Cache-Control"] = "public, max-age=36000"
        return response

    # /ws
    def _websocket(self, ws, endpoint= ""):
        endpoint = flask.request.cookies.get('endpoint',"")
        if flask.request.cookies.get('token') != self._token_cookie or flask.request.args.get('token') != self._token_csrf:
            msg = "token_failed"
        elif endpoint not in self._endpoints:
            msg = "no_endpoint"
        else:
            self._endpoints[endpoint].process_websocket(ws)
            msg = None
        try:
            ws.close(message=msg)
        except:
            pass

    def _add_file_to_monitor(self, file_to_monitor, class_name) -> None:
        if self.auto_reload is False:
            return
        if file_to_monitor not in self._file_monitoring:
            last_changed = os.path.getmtime(file_to_monitor)
            self._file_monitoring[file_to_monitor] = {
                "file_to_monitor": file_to_monitor,
                "last_changed": last_changed,
                "class_names": set(),
            }
        self._file_monitoring[file_to_monitor]["class_names"].add(class_name)

    def _monitoring_thread(self) -> None:
        while self.auto_reload is True:
            time.sleep(5)
            has_changed = False
            classed_to_reload = []
            for file_to_monitor in self._file_monitoring:
                data = self._file_monitoring[file_to_monitor]
                current_ts = os.path.getmtime(data["file_to_monitor"])
                if current_ts != data["last_changed"]:
                    self._file_monitoring[file_to_monitor]["last_changed"] = current_ts
                    classed_to_reload.extend(self._file_monitoring[file_to_monitor]["class_names"])
                    has_changed = True
            if has_changed is True:
                for endpoint in self._endpoints:
                    for instance in self._endpoints[endpoint]._gui_instances:
                        instance.clear_template_cache(classed_to_reload)
                        try:
                            instance.update()
                        except Exception:
                            logging.error("Failed to update, %s" % traceback.format_exc())


class WebsocketServerThread(threading.Thread):
    def __init__(self, app, host, port ):
        threading.Thread.__init__(self, daemon=True)
        self.server = make_server(host, port, app, threaded=True)
        self.ctx = app.app_context()
        self.ctx.push()

    def get_port(self):
        return self.server.server_port

    def run(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()


class PyHtmlGuiEndpoint:
    def __init__(self, parent: PyHtmlGui, app_instance: object, view_class: typing.Type[PyHtmlView], name: str = "",
         base_template    : str             = "pyHtmlGuiBase.html",
         on_view_connected: typing.Callable = None, on_view_disconnected: typing.Callable = None,
         single_instance : bool = True, size: tuple[int, int] = None, position: tuple[int, int] = None,
    ) -> None:
        self.parent = parent
        self.view_class = view_class
        self.app_instance = app_instance
        self.base_template = base_template
        self.size = size
        self.on_view_connected_callback = on_view_connected
        self.on_view_disconnected_callback = on_view_disconnected
        self.position = position
        self.single_instance = single_instance
        self.name = name
        self._gui_instances = []

    def process_websocket(self, ws):
        if len(self._gui_instances) == 0 or self.single_instance is False:
            instance = PyHtmlGuiInstance(self.parent, self.app_instance, self.view_class, self._on_dom_ready_callback)
            self._gui_instances.append(instance)
        else:
            instance = self._gui_instances[0]

        instance.process(ws)  # loop while connected

        # release/remove Instance
        if instance.connections_count == 0:
            self._gui_instances.remove(instance)
        if self.on_view_disconnected_callback is not None:
            self.on_view_disconnected_callback(len(self._gui_instances), sum([i.connections_count for i in self._gui_instances]))

    def _on_dom_ready_callback(self):
        if self.on_view_connected_callback is not None:
            self.on_view_connected_callback(len(self._gui_instances), sum([i.connections_count for i in self._gui_instances]))

    def get_url(self):
        target_host = "127.0.0.1" if self.parent.listen_host == "0.0.0.0" else self.parent.listen_host
        port = self.parent.listen_port if self.parent.listen_port != 0 else self.parent._server.get_port()
        url = "http://%s:%s/%s" % (target_host, port, self.name)
        if self.parent.shared_secret is not None:
            url = "%s?token=%s" % (url, self.parent.shared_secret)
        return url

    def show(self) -> None:
        webbrowser.open(self.get_url(), 1)



import shlex, os, sys
import subprocess
import tempfile
import threading
import logging
from pyhtmlgui import Observable


class PyHtmlChromeApp():
    def __init__(self, url, executable=None, commandline_args = [], ):
        self.url = url
        self.executable = executable
        self.commandline_args = commandline_args
        if self.executable == None:
            self.executable = self._find_path()
        if self.executable == None or not os.path.exists(self.executable):
            raise Exception("Chrome executable not found at path '%s'" % self.executable)
        self.on_closed_event = Observable()
        self.on_show_event = Observable()

        self._subthread = None
        self._subprocess = None

    def close(self):
        if self._subprocess is not None:
            self._subprocess.terminate()

    def show(self):
        if self._subthread == None:
            self._subthread = threading.Thread(target=self._chrome_thread, daemon=True)
            self._subthread.start()


    def join(self):
        if self._subprocess is not None:
            self._subprocess.wait()

    def _chrome_thread(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            cmd = [self.executable]
            cmd.extend(self.commandline_args)
            # cmd.append("--kiosk")
            cmd.append("--no-default-browser-check")
            cmd.append("--disable-logging")
            cmd.append("--content-shell-hide-toolbar")
            cmd.append("--bwsi")
            cmd.append("--no-service-autorun")
            cmd.append("--no-first-run")
            cmd.append("--no-report-upload")
            cmd.append("--overscroll-history-navigation=0")
            cmd.append("--disable-pinch")
            cmd.append("--user-data-dir=\"%s\"" % tmpdirname)
            cmd.append("--app=%s" % self.url)
            logging.info("Starting chrome instance with %s" % cmd)
            self._subprocess = subprocess.Popen(shlex.split(shlex.join(cmd)), shell=False)
            self.on_show_event.notify_observers()
            self._subprocess.wait()
        self._subthread = None
        self._subprocess = None
        self.on_closed_event.notify_observers()

    def _find_path(self):
        if sys.platform in ['win32', 'win64']:
            return self._find_chrome_win()
        elif sys.platform == 'darwin':
            return self._find_chrome_mac()
        elif sys.platform.startswith('linux'):
            return self._find_chrome_linux()
        else:
            return None

    @staticmethod
    def _find_chrome_mac():
        default_dir = r'/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
        if os.path.exists(default_dir):
            return default_dir
        # use mdfind ci to locate Chrome in alternate locations and return the first one
        name = 'Google Chrome.app'
        alternate_dirs = [x for x in subprocess.check_output(["mdfind", name]).decode().split('\n') if x.endswith(name)]
        if len(alternate_dirs):
            return alternate_dirs[0] + '/Contents/MacOS/Google Chrome'
        return None

    @staticmethod
    def _find_chrome_linux():
        import whichcraft as wch
        chrome_names = ['chromium-browser',
                        'chromium',
                        'google-chrome',
                        'google-chrome-stable']

        for name in chrome_names:
            chrome = wch.which(name)
            if chrome is not None:
                return chrome
        return None

    @staticmethod
    def _find_chrome_win():
        import winreg as reg
        reg_path = r'SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\chrome.exe'

        for install_type in reg.HKEY_CURRENT_USER, reg.HKEY_LOCAL_MACHINE:
            try:
                reg_key = reg.OpenKey(install_type, reg_path, 0, reg.KEY_READ)
                chrome_path = reg.QueryValue(reg_key, None)
                reg_key.Close()
            except WindowsError:
                chrome_path = None
            else:
                break

        return chrome_path
